import math
import os

class AnimationExporter:
	def __init__(self):
		self.export_all = False
		self.collection = True
		self.looping_threshold = 0.001

	def export_to_file(self, context, out_fn):
		if self.export_all:
			actions = []
			for o in context.selected_objects:
				if not o.animation_data:
					continue

				for t in o.animation_data.nla_tracks:
					for s in t.strips:
						if s.action and s.action not in actions:
							actions.append(s.action)

			if not actions:
				raise Exception("No actions found")

			resources = {}
			for a in actions:
				resources[a.name+".anim"] = self.export_animation(context, a)

			path, base = os.path.split(out_fn)
			base, ext = os.path.splitext(base)

			if self.collection:
				from .datafile import Statement
				with open(os.path.join(path, base+".mdc"), "w") as out_file:
					for r in resources.values():
						st = Statement("animation", r.name)
						st.sub = r.statements
						st.write_to_file(out_file)
			else:
				for r in resources.values():
					with open(os.path.join(path, r.name), w) as out_file:
						for s in r.statements:
							s.write_to_file(out_file)
		else:
			anim_data = context.active_object.animation_data
			if not anim_data:
				raise Exception("Object has no animation data")
			if not anim_data.action:
				raise Exception("No active action")

			resource = self.export_animation(context, anim_data.action)

			with open(out_fn, "w") as out_file:
				for s in resource.statements:
					s.write_to_file(out_file)

	def export_animation(self, context, action):
		from .animation import create_animation_from_action
		anim = create_animation_from_action(context, action, looping_threshold=self.looping_threshold)

		from .datafile import Resource, Statement
		resource = Resource(action.name+".anim")

		components = [(0, "location", "position"), (1, "rotation_euler", "euler"), (2, "scale", "scale")]
		coords = "xyz"
		prev_time = 0.0
		for k in anim.keyframes:
			if k.time>prev_time:
				resource.statements.append(Statement("interval", k.time-prev_time))
				prev_time = k.time

			st = Statement("control_keyframe" if k.control else "keyframe")

			transform = [0.0]*9
			mask = 0
			for c, i in k.curves:
				for j, dp, kw in components:
					if c.data_path==dp:
						transform[j*3+c.array_index] = c.knots[i][1]
						mask |= 1<<(j*3+c.array_index)
						break

			if mask:
				ss = Statement("transform")

				for i, dp, kw in components:
					v = transform[i*3:i*3+3]
					if i==1:
						v = [c*180/math.pi for c in v]

					m = 7<<(i*3)
					if (mask&m)==m:
						ss.sub.append(Statement(kw, *v))
					else:
						m &= m>>2
						for j in range(3):
							if mask&(m<<j):
								ss.sub.append(Statement("{}_{}".format(kw, coords[j]), v[j]))

				st.sub.append(ss)

			resource.statements.append(st)

		resource.statements.append(Statement("looping", anim.looping))

		return resource

