#include <msp/gl/extensions/arb_vertex_array_object.h>
#include <msp/gl/extensions/arb_vertex_buffer_object.h>
#include <msp/gl/extensions/arb_vertex_shader.h>
#include <msp/gl/extensions/nv_primitive_restart.h>
#include "buffer.h"
#include "error.h"
#include "mesh.h"
#include "renderer.h"
#include "resourcemanager.h"
#include "vertexsetup.h"

using namespace std;

namespace Msp {
namespace GL {

Mesh::Mesh(ResourceManager *rm):
	vertices(VERTEX3)
{
	init(rm);
}

Mesh::Mesh(const VertexFormat &f, ResourceManager *rm):
	vertices(f)
{
	init(rm);
}

void Mesh::init(ResourceManager *rm)
{
	vbuf = 0;
	ibuf = 0;
	vtx_setup = 0;
	defer_buffers = ARB_vertex_buffer_object;
	disallow_rendering = false;
	winding = 0;

	if(rm)
		set_manager(rm);
}

Mesh::~Mesh()
{
	set_manager(0);
	delete vtx_setup;
	delete vbuf;
	delete ibuf;
}

void Mesh::clear()
{
	vertices.clear();
	batches.clear();
}

void Mesh::use_buffers(bool b)
{
	defer_buffers = false;
	if(b)
		create_buffers();
	else
	{
		vertices.use_buffer(0);
		delete vbuf;
		vbuf = 0;
		delete ibuf;
		ibuf = 0;
	}
}

void Mesh::create_buffers()
{
	defer_buffers = false;

	if(!vbuf)
		vbuf = new Buffer(ARRAY_BUFFER);
	vertices.use_buffer(vbuf);

	if(!ibuf)
		ibuf = new Buffer(ELEMENT_ARRAY_BUFFER);

	if(ARB_vertex_array_object && !vtx_setup)
	{
		vtx_setup = new VertexSetup;
		vtx_setup->set_vertex_array(vertices);
		vtx_setup->set_index_buffer(*ibuf);
	}
}

unsigned Mesh::get_n_vertices() const
{
	return vertices.size();
}

float *Mesh::modify_vertex(unsigned i)
{
	return vertices.modify(i);
}

void Mesh::add_batch(const Batch &b)
{
	if(defer_buffers)
		create_buffers();

	if(!batches.empty() && batches.back().can_append(b.get_type()))
		batches.back().append(b);
	else
	{
		bool reallocate = (batches.size()==batches.capacity());
		if(reallocate && ibuf)
		{
			for(vector<Batch>::iterator i=batches.end(); i!=batches.begin(); )
				(--i)->use_buffer(0);
		}

		Batch *prev = (batches.empty() ? 0 : &batches.back());
		batches.push_back(b);
		if(ibuf)
		{
			if(reallocate)
			{
				prev = 0;
				for(vector<Batch>::iterator i=batches.begin(); i!=batches.end(); ++i)
				{
					i->use_buffer(ibuf, prev);
					prev = &*i;
				}
			}
			else
				batches.back().use_buffer(ibuf, prev);
		}
	}
}

void Mesh::set_winding(const WindingTest *w)
{
	winding = w;
}

void Mesh::draw() const
{
	const Mesh *cur = current();
	if(cur && cur!=this)
		throw invalid_operation("Mesh::draw");

	if(manager)
	{
		manager->resource_used(*this);
		if(disallow_rendering)
			return;
	}

	if(!current())
		vertices.apply();
	BindRestore bind_ibuf(ibuf, ELEMENT_ARRAY_BUFFER);
	Bind bind_winding(winding);

	for(vector<Batch>::const_iterator i=batches.begin(); i!=batches.end(); ++i)
		i->draw();
}

void Mesh::draw(Renderer &renderer) const
{
	if(manager)
	{
		manager->resource_used(*this);
		if(disallow_rendering)
			return;
	}

	renderer.set_mesh(this);
	renderer.set_winding_test(winding);

	for(vector<Batch>::const_iterator i=batches.begin(); i!=batches.end(); ++i)
		renderer.draw(*i);
}

void Mesh::bind() const
{
	/* If VAOs are not supported, vtx_setup is zero and set_current won't get
	called.  Thus unbind won't try to call a null function either. */
	if(!vtx_setup)
	{
		unbind();
		vertices.apply(false);
	}
	else if(set_current(this))
	{
		vtx_setup->bind();
		vertices.refresh();
	}
}

void Mesh::unbind()
{
	if(set_current(0))
		VertexSetup::unbind();
}

Resource::AsyncLoader *Mesh::load(IO::Seekable &io, const Resources *)
{
	return new AsyncLoader(*this, io);
}

UInt64 Mesh::get_data_size() const
{
	UInt64 size = 0;
	if(vbuf)
		size += vbuf->get_size();
	if(ibuf)
		size += ibuf->get_size();
	return size;
}

void Mesh::unload()
{
	vertices.clear();
	vertices.use_buffer(0);
	batches.clear();
	delete vbuf;
	delete ibuf;
	defer_buffers = (vbuf || ibuf);
	vbuf = 0;
	ibuf = 0;
}


Mesh::Loader::Loader(Mesh &m):
	DataFile::ObjectLoader<Mesh>(m)
{
	add("batch",    &Loader::batch);
	add("vertices", &Loader::vertices);
	add("winding",  &Loader::winding);
}

void Mesh::Loader::vertices(const vector<VertexComponent> &c)
{
	if(c.empty())
		throw invalid_argument("No vertex components");

	VertexFormat fmt;
	for(vector<VertexComponent>::const_iterator i=c.begin(); i!=c.end(); ++i)
		fmt = (fmt, *i);
	obj.vertices.reset(fmt);
	if(obj.vtx_setup)
		// Set it again to force the vertex setup to update
		obj.vtx_setup->set_vertex_array(obj.vertices);
	load_sub(obj.vertices);
}

void Mesh::Loader::batch(PrimitiveType p)
{
	Batch btc(p);
	load_sub(btc);
	obj.add_batch(btc);
}

void Mesh::Loader::winding(FaceWinding w)
{
	if(w==CLOCKWISE)
		obj.winding = &WindingTest::clockwise();
	else if(w==COUNTERCLOCKWISE)
		obj.winding = &WindingTest::counterclockwise();
}


Mesh::AsyncLoader::AsyncLoader(Mesh &m, IO::Seekable &i):
	mesh(m),
	io(i),
	vertex_updater(0),
	index_updater(0),
	phase(0)
{
	// Make sure the extension is initialized in the rendering thread.
	(void)(bool)NV_primitive_restart;

	mesh.disallow_rendering = true;
	if(mesh.defer_buffers)
		mesh.create_buffers();
}

Mesh::AsyncLoader::~AsyncLoader()
{
	mesh.disallow_rendering = false;
	delete vertex_updater;
	delete index_updater;
}

bool Mesh::AsyncLoader::needs_sync() const
{
	return phase%2;
}

bool Mesh::AsyncLoader::process()
{
	if(phase==0)
	{
		// TODO use correct filename
		DataFile::Parser parser(io, "async");
		Loader loader(mesh);
		loader.load(parser);
	}
	else if(phase==1)
	{
		vertex_updater = mesh.vertices.refresh_async();
		if(!mesh.batches.empty())
			index_updater = mesh.batches.front().refresh_async();
	}
	else if(phase==2)
	{
		if(vertex_updater)
			vertex_updater->upload_data();
		if(index_updater)
			index_updater->upload_data();
	}
	else if(phase==3)
	{
		delete vertex_updater;
		vertex_updater = 0;
		delete index_updater;
		index_updater = 0;
	}

	++phase;
	if(phase==1 && !mesh.vbuf && !mesh.ibuf)
		phase += 3;
	return phase>3;
}

} // namespace GL
} // namespace Msp
